package routing;

import core.*;
import interfaces.EnergyInterface;
import org.apache.commons.lang3.tuple.Pair;
import routing.util.EnergyModule;
import routing.util.EnergySeeker;

import java.util.*;

public abstract class BasicEnergyRouter extends ActiveRouter {

    protected static final String ROUTER_PREFIX = "BasicEnergyRouter.";
    protected static final String EXCHANGE_PROPERTY_RESTRICTIONS =
            ROUTER_PREFIX + "restrictions";
    protected static final String EXCHANGE_PROPERTY_STATE =
            ROUTER_PREFIX + "state";
    protected static final String EXCHANGE_PROPERTY_IDENTITY =
            ROUTER_PREFIX + "id";
    protected static final String EXCHANGE_PROPERTY_QUEUE =
            ROUTER_PREFIX + "queue";
    protected static final String EXCHANGE_PROPERTY_SUMMARY =
            ROUTER_PREFIX + "summary";

    // enum for state status
    protected enum State {
        SEEK, SUPPLY, NEUTRAL
    }

    protected static int protocolMsgId = 0;

    protected HashMap<DTNHost, Contact> contacts = new HashMap<>();
    protected List<Contact> suppliers = new LinkedList<>();
    protected List<Contact> seekers = new LinkedList<>();
    protected double lastUpdateTime = 0;
    protected State state;
    protected String identity;
    protected Set<String> restrictions;

    protected int plugCount;
    protected double price;

    public BasicEnergyRouter(Settings s) {
        super(s);

        this.plugCount = s.getInt("plugCount", 1);
        this.price = s.getInt("price", 1);
    }
    public BasicEnergyRouter(BasicEnergyRouter prototype) {
        super(prototype);
        this.plugCount = prototype.plugCount;
        this.price = prototype.price;
    }

    @Override
    public void changedConnection(Connection con) {
        super.changedConnection(con);
        DTNHost thisHost = getHost();
        DTNHost peer = con.getOtherNode(thisHost);
        Pair<NetworkInterface, NetworkInterface> iFaces = con.getInterfaces();

        if (iFaces.getLeft() instanceof EnergyInterface &&
                iFaces.getRight() instanceof EnergyInterface) {
            if (con.isUp()) {

                Contact c = new Contact();
                c.host = peer;
                this.contacts.put(peer, c);
            } else {
                this.contacts.remove(peer);
                this.suppliers.removeIf(c -> c.host == peer);
                this.seekers.removeIf(c -> c.host == peer);
            }
        }
    }

    @Override
    public Message messageTransferred(String id, DTNHost from) {
        Message m = super.messageTransferred(id, from);
        State peerState;
        Contact contact;
        String identity;
        if (!this.isDeliveredMessage(m)) {
            return m;
        }
        peerState = (State) m.getProperty(EXCHANGE_PROPERTY_STATE);
        this.seekers.removeIf(c -> c.host == from);
        this.suppliers.removeIf(c -> c.host == from);
        switch (peerState) {
            case SEEK:
                if (this.state == State.SUPPLY) {
                    this.seekers.removeIf(c -> c.host == from);
                    identity = (String) m.getProperty
                            (EXCHANGE_PROPERTY_IDENTITY);
                    contact = new Contact();
                    contact.state = peerState;
                    contact.identity = identity;
                    contact.host = from;
                    this.seekers.add(contact);
                    this.contacts.put(from, contact);
                }
                break;

            case SUPPLY:
                if (this.state == State.SEEK) {
                    this.suppliers.removeIf(c -> c.host == from);
                    @SuppressWarnings("unchecked") Set<String> restrictions
                            = (Set<String>) m.getProperty
                            (EXCHANGE_PROPERTY_RESTRICTIONS);
                    identity = (String) m.getProperty
                            (EXCHANGE_PROPERTY_IDENTITY);
                    @SuppressWarnings("unchecked") List<Pair<String,
                            Double>> queue = (List<Pair<String, Double>>) m
                            .getProperty(EXCHANGE_PROPERTY_QUEUE);
                    contact = new Contact();
                    contact.state = peerState;
                    contact.restrictions = restrictions;
                    contact.identity = identity;
                    contact.host = from;
                    contact.queue = queue;
                    this.suppliers.add(contact);
                    this.contacts.put(from, contact);
                }
                break;
            case NEUTRAL:
                if (this.state == State.SUPPLY) {
                    this.seekers.removeIf(c -> c.host == from);
                }
            default:
                break;
        }

        return m;
    }

    public void update() {
        super.update();
        double time = SimClock.getTime();
        double interval = time - this.lastUpdateTime;
        this.lastUpdateTime = time;
        Message message;
        this.updateState();
        switch (this.state) {
            case SEEK:
                if (!this.suppliers.isEmpty()) {
                    Contact supplier = this.pickDesiredSupplier();
                    message = new Message(this.getHost(), supplier.host,
                            EXCHANGE_PROPERTY_SUMMARY + protocolMsgId++, 64);
                    message.addProperty(EXCHANGE_PROPERTY_IDENTITY, this
                            .identity);
                    message.addProperty(EXCHANGE_PROPERTY_STATE, this.state);
                    createNewMessage(message);
                    for (int i = 1; i < this.suppliers.size(); i++) {
                        supplier = this.suppliers.get(i);
                        message = new Message(this.getHost(), supplier.host,
                                EXCHANGE_PROPERTY_SUMMARY +
                                        protocolMsgId++, 64);
                        message.addProperty(EXCHANGE_PROPERTY_IDENTITY, this
                                .identity);
                        message.addProperty(EXCHANGE_PROPERTY_STATE, State
                                .NEUTRAL);
                        createNewMessage(message);
                    }
                }
                break;

            case SUPPLY:
                for (int i = 0;
                     i < this.seekers.size() && i < this.plugCount; i++) {
                    MessageRouter r = this.seekers.get(i).host.getRouter();
                    if (r instanceof EnergySeeker) {
                        this.chargePeer((EnergySeeker) r, interval);
                    } else {
                        System.out.println("BasicEnergyRouter: failed");
                        System.exit(1);
                    }
                }
                for (Map.Entry<DTNHost, Contact> entry : this.contacts
                        .entrySet()) {
                    Contact contact = entry.getValue();
                    //System.out.println("Supplier sending message...");
                    Message m = new Message(this.getHost(), contact.host,
                            EXCHANGE_PROPERTY_SUMMARY + protocolMsgId++, 64);
                    m.addProperty(EXCHANGE_PROPERTY_IDENTITY, this.identity);
                    m.addProperty(EXCHANGE_PROPERTY_STATE, this.state);
                    m.addProperty(EXCHANGE_PROPERTY_QUEUE, this.seekers);
                    m.addProperty(EXCHANGE_PROPERTY_RESTRICTIONS, this
                            .restrictions);
                    createNewMessage(m);

                }
                break;

            case NEUTRAL:
            default:
                break;
        }
        if (isTransferring() || !canStartTransfer()) {
            return;
        }
        exchangeDeliverableMessages();
    }

    protected void updateState() {
        EnergyModule em = this.getEnergyModule();
        switch (this.state) {
            case NEUTRAL:
                if (em.getRemainingCapacity() < 100) {
                    this.state = State.SEEK;
                }
            case SEEK:
                if (em.getRemainingCapacity() == 100) {
                    this.state = State.NEUTRAL;
                }
            case SUPPLY:
            default:
        }
    }
    protected abstract void chargePeer(EnergySeeker seeker, double interval);
    public double getPrice() {
        return price;
    }
    protected abstract EnergyModule getEnergyModule();
    protected Contact pickDesiredSupplier() {
        return this.suppliers.get(0);
    }
    protected class Contact {
        State state;
        Set<String> restrictions;
        String identity;
        List<Pair<String, Double>> queue;
        DTNHost host;
        double capacity;
    }
}
