package routing;
import core.*;
import interfaces.EnergyInterface;
import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;
import routing.util.EnergyModule;
import routing.util.EnergySeeker;
import java.util.*;
import java.util.stream.Collectors;
public abstract class EnergyRouter extends ActiveRouter {
    private static final String ROUTER_PREFIX = "EnergyRouter.";
    protected static final String EXCHANGE_PROPERTY_STATE =
            ROUTER_PREFIX + "state";
    protected static final String EXCHANGE_PROPERTY_RESTRICTIONS =
            ROUTER_PREFIX + "restrictions";
    protected static final String EXCHANGE_PROPERTY_IDENTITY =
            ROUTER_PREFIX + "id";
    protected static final String EXCHANGE_PROPERTY_QUEUE =
            ROUTER_PREFIX + "queue";
    protected static final String EXCHANGE_PROPERTY_SUMMARY =
            ROUTER_PREFIX + "summary";
    protected static final String EXCHANGE_PROPERTY_CAPACITY =
            ROUTER_PREFIX + "capacity";
    protected static final String EXCHANGE_PROPERTY_PLUG_COUNT =
            ROUTER_PREFIX + "plugCount";
    protected enum State {
        SEEK, SUPPLY, NEUTRAL
    }
    private static int protocolMsgId = 0;
    private HashMap<DTNHost, Contact> contacts = new HashMap<>();
    private double lastUpdateTime = 0;
    private String identity = "";
    protected int plugCount;
    private double price;
    private Set<String> restrictions = new HashSet<>();
    private double lastHello = 0.0;
    protected State state;
    protected EnergyModule energyModule;
    protected List<Contact> suppliers = new LinkedList<>();
    protected List<Contact> seekers = new LinkedList<>();
    public EnergyRouter(Settings s) {
        super(s);
       this.plugCount = s.getInt("plugCount", 1);
        this.price = s.getInt("price", 1);
    }
    publicEnergyRouter(EnergyRouter prototype) {
        super(prototype);
        this.plugCount = prototype.plugCount;
        this.price = prototype.price;
    }
    @Override
    public abstract MessageRouter replicate();
    @Override
    public void changedConnection(Connection con) {
        super.changedConnection(con);
        DTNHost thisHost = getHost();
        DTNHost peer = con.getOtherNode(thisHost);
        Pair<NetworkInterface, NetworkInterface> iFaces = con.getInterfaces();
        if (iFaces.getLeft() instanceof EnergyInterface &&
                iFaces.getRight() instanceof EnergyInterface) {
            if (con.isUp()) {
                Contact c = new Contact();
                c.host = peer;
                c.connectionStartTime = SimClock.getTime();
                this.contacts.put(peer, c);
            } else {
                this.contacts.remove(peer);
                this.suppliers.removeIf(c -> c.host == peer);
                this.seekers.removeIf(c -> c.host == peer);
            }
        }
    }

    @Override
    public Message messageTransferred(String id, DTNHost from) {
        Message m = super.messageTransferred(id, from);
        State peerState;
        Contact contact;
        if (this.isDeliveredMessage(m)) {
            peerState = (State) m.getProperty(EXCHANGE_PROPERTY_STATE);
            this.seekers.removeIf(c -> c.host == from);
            this.suppliers.removeIf(c -> c.host == from);
            contact = this.contacts.get(from);
            if (contact == null) {
                contact = new Contact();
                contact.host = from;
                contact.connectionStartTime = SimClock.getTime();
                contact.state = peerState;
                this.contacts.put(from, contact);
            }
            switch (peerState) {
                case SEEK:
                    if (this.state == State.SUPPLY) {
                        contact.identity = (String) m.getProperty
                                (EXCHANGE_PROPERTY_IDENTITY);
                        contact.capacity = (double) m.getProperty
                                (EXCHANGE_PROPERTY_CAPACITY);
                        this.seekers.add(contact);
                    }
                    break;

                case SUPPLY:
                    if (this.state == State.SEEK) {
                        @SuppressWarnings("unchecked") Set<String>
                                restrictions = (Set<String>) m.getProperty
                                (EXCHANGE_PROPERTY_RESTRICTIONS);
                        @SuppressWarnings("unchecked") List<Pair<String,
                                Double>> queue = (List<Pair<String,
                                Double>>) m.getProperty
                                (EXCHANGE_PROPERTY_QUEUE);
                        contact.restrictions = restrictions;
                        contact.identity = (String) m.getProperty
                                (EXCHANGE_PROPERTY_IDENTITY);
                        contact.queue = queue;
                        contact.capacity = (double) m.getProperty
                                (EXCHANGE_PROPERTY_CAPACITY);
                        contact.plugCount = (int) m.getProperty
                                (EXCHANGE_PROPERTY_PLUG_COUNT);
                        this.suppliers.add(contact);
                    }
                    break;
                case NEUTRAL:
                default:
                    break;
            }
        }

        return m;
    }

    @Override
    public void update() {
        super.update();
        double time = SimClock.getTime();
        double interval = time - this.lastUpdateTime
        exchangeDeliverableMessages();
        if (interval < 600) {
            return;
        }
        this.lastUpdateTime = time;
        Message message;
        this.updateState();
        switch (this.state) {
            case SEEK:
                if (!this.suppliers.isEmpty()) {
                    Contact supplier = this.pickDesiredSupplier();
                    if (supplier == null) {
                        break;
                    }
                    if (!this.contacts.containsKey(supplier.host)) {
                        System.out.println("Attempting to send to supplier who is not a contact!");
                        System.exit(1);
                    }
                    message = new Message(this.getHost(), supplier.host,
                            EXCHANGE_PROPERTY_SUMMARY + protocolMsgId++, 64);
                    message.addProperty(EXCHANGE_PROPERTY_IDENTITY, this
                            .identity);
                    message.addProperty(EXCHANGE_PROPERTY_STATE, this.state);
                    message.addProperty(EXCHANGE_PROPERTY_CAPACITY, this
                            .energyModule.getRemainingCapacity());
                    createNewMessage(message);
                }
                break;

            case SUPPLY
                this.orderSeekers();
                for (int i = 0;
                     i < this.seekers.size() && i < this.plugCount; i++) {
                    MessageRouter r = this.seekers.get(i).host.getRouter();
                    if (r instanceof EnergySeeker) {
                        this.chargePeer((EnergySeeker) r, interval);
                    } else {
                        System.out.println("EnergyRouter: attempting to" +
                                " charge a router which is not an " +
                                "EnergySeeker");
                        System.exit(1);
                    }
                }
                List<Pair<String, Double>> suppliedQueue = this.seekers
                        .stream().map(c -> new ImmutablePair<>(c.identity, c
                                .capacity)).collect(Collectors.toList());
                for (Map.Entry<DTNHost, Contact> entry : this.contacts
                        .entrySet()) {
                    Contact contact = entry.getValue();
                    Message m = new Message(this.getHost(), contact.host,
                            EXCHANGE_PROPERTY_SUMMARY + protocolMsgId++, 64);
                    m.addProperty(EXCHANGE_PROPERTY_IDENTITY, this.identity);
                    m.addProperty(EXCHANGE_PROPERTY_STATE, this.state);
                    m.addProperty(EXCHANGE_PROPERTY_QUEUE, suppliedQueue);
                    m.addProperty(EXCHANGE_PROPERTY_RESTRICTIONS, this
                            .restrictions);
                    m.addProperty(EXCHANGE_PROPERTY_CAPACITY, this
                            .energyModule.getRemainingCapacity());
                    m.addProperty(EXCHANGE_PROPERTY_PLUG_COUNT, this
                            .plugCount);
                    createNewMessage(m);
                }
                break;
            case NEUTRAL:
            default:
                break;
        }
        this.suppliers.clear();
        this.seekers.clear();
    }

    @Override
    protected void transferDone(Connection con) {
        String msgId = con.getMessage().getId();
        Message msg = getMessage(msgId);

        if (msg != null &&
                msg.getProperty(EXCHANGE_PROPERTY_SUMMARY) != null) {
            deleteMessage(msgId, false);
        }
    }
    protected abstract void updateState();
    protected abstract Contact pickDesiredSupplier();
    protected abstract void orderSeekers();
    protected abstract void chargePeer(EnergySeeker seeker, double interval);
    class Contact {
        State state;
        Set<String> restrictions;
        String identity;
        List<Pair<String, Double>> queue;
        DTNHost host;
        double capacity;
        double connectionStartTime;
        int plugCount;
    }
}
